#  ---------------------------------------------------------------------
#   General database implementation of ODL-Thread (v1.0) - data provider
#    v1.0
#    August 2002
#  ------------------+--------------------+-----------------------------
#   Hussein Suleman  |   hussein@vt.edu   |    www.husseinsspace.com    
#  ------------------+--------------------+-+---------------------------
#   Department of Computer Science          |        www.cs.vt.edu       
#     Digital Library Research Laboratory   |       www.dlib.vt.edu      
#  -----------------------------------------+-------------+-------------
#   Virginia Polytechnic Institute and State University   |  www.vt.edu  
#  -------------------------------------------------------+-------------


package ODL::Thread::ThreadDP;


use POSIX;
use DBI;

use Pure::EZXML;
use Pure::X2D;
use ODL::Thread::ThreadDB;

use XOAI::XOAIDP;
use vars ('@ISA');
@ISA = ("XOAI::XOAIDP");


# constructor
sub new
{
   my ($classname, $configfile) = @_;
   
   my $self = $classname->SUPER::new ();

   # get configuration from file
   my $con = new Pure::X2D ($configfile);
   $self->{'database'} = $con->param ('database', '');
   $self->{'dbusername'} = $con->param ('dbusername', '');
   $self->{'dbpassword'} = $con->param ('dbpassword', '');
   $self->{'table'} = $con->param ('table', 'irdb');
   $self->{'repositoryName'} = $con->param ('repositoryName', 'ODL Box Archive');
   $self->{'adminEmail'} = $con->param ('adminEmail', "mailto:someone\@somewhere");
   $self->{'recordlimit'} = $con->param ('recordlimit', 100);
   $self->{'archiveId'} = $con->param ('archiveId', 'box');
   
   # set internal variables
   $self->{'resumptionseparator'} = '!';
   $self->{'threadepoch'} = 20;
   
   # open database
   $self->{'dbh'} = ODL::Thread::ThreadDB::openDatabase ($self->{'database'},
     $self->{'dbusername'}, $self->{'dbpassword'}, $self->{'table'});
     
   $self->{'odlsearch_xmlnsprefix'} = 'http://oai.dlib.vt.edu/ODL/Search/1.0';
   $self->{'odlannotate_xmlnsprefix'} = 'http://oai.dlib.vt.edu/ODL/Annotate/1.0';

   bless $self, $classname;
   return $self;
}


# destructor
sub dispose
{
   my ($self) = @_;
   $self->SUPER::dispose ();
   $self->{'dbh'}->disconnect;
}


# convert unix calendar time to ISO8601 format
sub toISO8601
{
   my ($self, $atime, $granularity) = @_;
   
   my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime ($atime);
   if ((defined $granularity) && ($granularity eq 'second'))
   {
      sprintf ("%04d-%02d-%02dT%02d:%02d:%02d+00:00", $year+1900, $mon+1, $mday, $hour, $min, $sec);
   }
   else
   {
      sprintf ("%04d-%02d-%02d", $year+1900, $mon+1, $mday);
   }
}


# convert an ISO8601 date into a unix calendar time representation
sub toSeconds
{
   my ($self, $date, $from) = @_;
   
   my ($hour, $minute, $second);
   
   if ((defined $from) && ($from == 1))
   {
      ($hour, $minute, $second) = (0, 0, 0);
   }
   else
   {
      ($hour, $minute, $second) = (23, 59, 59);
   }
   
   if ($date =~ /([0-9]{4})-([0-9]{2})-([0-9]{2})T([0-9]{2}):([0-9]{2}):([0-9]{2})[\+\-][0-9]{2}:[0-9]{2}/)
   {
      return mktime ($6, $5, $4, $3, $2-1, $1-1900);
   }
   elsif ($date =~ /([0-9]{4})-([0-9]{2})-([0-9]{2})/)
   {
      return mktime ($second, $minute, $hour, $3, $2-1, $1-1900);
   }
   return 0;
}

# component-specific identification
sub Archive_Identify
{
   my ($self) = @_;
   
   my $id = $self->SUPER::Archive_Identify ();
      
   if (! exists $id->{'description'})
   {
      $id->{'description'} = [];
   }
   my $desc = {
      'odl-description' => [[ 
         {
            'xmlns' => $self->{xoai_xmlnsprefix}.'/odl-description',
            'xsi:schemaLocation' => 
                       $self->{xoai_xmlnsprefix}.'/odl-description '.
                       $self->{xoai_xmlnsprefix}.'/odl-description.xsd'
         },
         {
            'protocol' => 'ODL-Annotate',
            'version'  => '1.0',
            'mdorder'  => [ qw ( protocol version ) ]
         } 
      ]]
   };
   push (@{$id->{'description'}}, $desc);
   
   $id;
}


# retrieve records from the source archive as required
sub Archive_FormatRecord
{
   my ($self, $hashref, $metadataFormat) = @_;
   
   if (defined $hashref->{$metadataFormat})
   {
      my $rowref = $hashref->{$metadataFormat};
      $self->FormatRecord ($rowref->{'identifier'},
                           $self->toISO8601 ($rowref->{'datestamp'}, 'second'),
                           $rowref->{'metadata'},
                           '<level '.
                            "xmlns=\"$self->{'odlannotate_xmlnsprefix'}/level\" ".
                            "xsi:schemaLocation=\"$self->{'odlannotate_xmlnsprefix'}/level ".
                            "$self->{'odlannotate_xmlnsprefix'}/level.xsd\">".
                            $rowref->{'level'}.
                           '</level>',
                           $rowref->{'status'}
                          );
   }
   elsif ((keys %$hashref) == 0)
   {
      return '';
   }
   else
   {
      my $firstkey = (keys %$hashref)[0];
      my $rowref = $hashref->{$firstkey};
      $self->FormatRecord ($rowref->{'identifier'},
                           $self->toISO8601 ($rowref->{'datestamp'}, 'second'),
                           '',
                           '',
                           $rowref->{'status'}
                          );
   }
}


# get full list of mdps or list for specific identifier
sub Archive_ListMetadataFormats
{
   my ($self, $identifier) = @_;
   
   my %metadataNamespace = ();
   my %metadataSchema = ();
   my @mdps = ();
   
   # get full list of metadataPrefixes
   my $sth;
   if ($identifier ne '')
   {
      $sth = $self->{'dbh'}->prepare ("select metadataPrefix from $self->{'table'} ".
                                      "where identifier=\'$identifier\'");
   }
   else
   {
      $sth = $self->{'dbh'}->prepare ("select metadataPrefix from $self->{'table'}md");
   }
   $sth->execute;
   while (my $rowref = $sth->fetchrow_hashref)
   {
      push (@mdps, $rowref->{'metadataPrefix'});
   }
   $sth->finish;
   if ($#mdps == -1)
   {
      return [];
   }
   
   # get namespaces and schemas
   foreach my $mdp (@mdps)
   {
      $sth = $self->{'dbh'}->prepare ("select metadataNamespace,metadataSchema ".
                                      "from $self->{'table'}md ".
                                      "where metadataPrefix=\'$mdp\'");
      $sth->execute;
      if (my $rowref = $sth->fetchrow_hashref)
      {
         $metadataNamespace{$mdp} = $rowref->{'metadataNamespace'};
         $metadataSchema{$mdp} = $rowref->{'metadataSchema'};
      }
   }

   [ \%metadataNamespace, \%metadataSchema];
}


# generate list of sets from database
sub Archive_ListSets
{
   my ($self) = @_;
   
   my $setlist = [];
      
   my $sth = $self->{'dbh'}->prepare ("select distinct oaiset from $self->{'table'}set");
   $sth->execute;
               
   my $rowref;
   while ($rowref = $sth->fetchrow_hashref)
   {
      push (@{$setlist}, [ $rowref->{'oaiset'}, $rowref->{'oaiset'} ]);
   }
                                 
   $sth->finish;
                                       
   $setlist;
}


# pass request through to source
sub Archive_GetRecord
{
   my ($self, $identifier, $metadataPrefix) = @_;
   
   my %records = ();
      
   if ($identifier =~ /^parent\/(.*)/)
   {
      my $child = $1;
      my $sth = $self->{'dbh'}->prepare ("select parent from $self->{'table'} ".
                                         "where identifier=\'$child\'");
      $sth->execute;
      my $rowref = $sth->fetchrow_hashref;
      $sth->finish;
      
      if (! $rowref)
      {
         return \%records;
      }
      $identifier = $rowref->{'parent'};
   }
   
   my $sth = $self->{'dbh'}->prepare ("select identifier,status,datestamp,metadataPrefix,".
                                      "metadata,level from $self->{'table'} ".
                                      "where identifier=\'$identifier\' and ".
                                      "metadataPrefix=\'$metadataPrefix\'");
   $sth->execute;

   my $rowref;
   while ($rowref = $sth->fetchrow_hashref)
   {
      $records{$rowref->{'metadataPrefix'}} = {
         identifier => $rowref->{'identifier'},
         datestamp  => $rowref->{'datestamp'},
         status     => $rowref->{'status'},
         metadata   => $rowref->{'metadata'},
         level      => $rowref->{'level'}
      };
   }
   
   $sth->finish;
                                                                                          
   \%records;
}


# return list of records/identifiers
sub Archive_ListRecords
{
   my ($self, $set, $from, $until, $metadataPrefix, $resumptionToken) = @_;

   my ($offset);
   if ($resumptionToken eq '')
   {
      $offset = 0;
   }
   else
   {
      my @rdata = split ($self->{resumptionseparator}, $resumptionToken);
      ($set, $from, $until, $metadataPrefix, $offset) = @rdata;
   }

   my $sth;

   # construct query
   my $sqlquery = "from $self->{'table'}";
   my $sqlquery1 = '';
   if ($metadataPrefix ne '') 
   { $sqlquery1 = "select * "; }
   else
   { $sqlquery1 = "select distinct status,identifier "; }

   if ($set ne '')
   {
      $sqlquery .= ",$self->{'table'}set";
   }

   if ($from eq '')
   {
      if ($until eq '') 
      { }
      else
      { $sqlquery = $sqlquery." where (datestamp <= \'".$self->toSeconds ($until)."\')"; }
   }
   else
   { 
      if ($until eq '')
      { $sqlquery = $sqlquery." where (datestamp >= \'".$self->toSeconds ($from, 1)."\')"; }
      else
      { $sqlquery = $sqlquery." where (datestamp >= \'".$self->toSeconds ($from, 1)."\') and (datestamp <= \'".$self->toSeconds ($until)."\')"; }
   }
   
   if ($metadataPrefix ne '')
   {
      if (($from eq '') && ($until eq ''))
      { $sqlquery .= ' where '; }
      else 
      { $sqlquery .= ' and '; }
      $sqlquery .= "($self->{'table'}.metadataPrefix = \'$metadataPrefix\')";
   }

   my $limitstr = '';
   my $explicitlimits = 0;
   if ($set ne '')
   {
      if (($from eq '') && ($until eq '') && ($metadataPrefix eq ''))
      { $sqlquery .= ' where '; }
      else 
      { $sqlquery .= ' and '; }
      
      if ($set =~ /^([0-9]+)\/([0-9]+)\/(.*)/)
      {
         my $limit = $2-$1+1;
         my $start = $1-1;
         $limitstr = " limit $start,$limit";
         $explicitlimits = 1;
         $set = $3;
      }
      
      $sqlquery .= "($self->{'table'}set.identifier=$self->{'table'}.identifier".
                   " and $self->{'table'}set.oaiset = \'$set\')";
   }
   
   if (($from ne '') || ($until ne ''))
   {
      $sqlquery .= " order by $self->{'table'}.datestamp desc";
   }
   else
   {
      $sqlquery .= " order by $self->{'table'}.thread desc";
   }

   if ($explicitlimits == 0)
   {
      my $limit = $self->{'recordlimit'} + 1;
      $limitstr = " limit $offset,$limit";
   }
   
   $sth = $self->{'dbh'}->prepare ($sqlquery1.$sqlquery.$limitstr);
   $sth->execute;

   my @allrows = ();
   my $gotmore = 0;
   my $count = 0; 

   while (my $rowref = $sth->fetchrow_hashref)
   {
      if (($count == $self->{'recordlimit'}) && ($explicitlimits == 0))
      { $gotmore = 1; }
      else
      {
         $count++;
         if ($metadataPrefix eq '')
         {
            push (@allrows, { 
               urn    => $rowref->{'identifier'},
               status => $rowref->{'status'}
            });
         }
         else
         {
            push (@allrows, {
               $metadataPrefix => {
                  identifier => $rowref->{'identifier'},
                  datestamp  => $rowref->{'datestamp'},
                  status     => $rowref->{'status'},
                  metadata   => $rowref->{'metadata'},
                  level      => $rowref->{'level'}
               }
            });
         }
      }
   }

   $sth->finish;

   $resumptionToken = '';
   if (($gotmore == 1) && ($explicitlimits == 0))
   {
      $resumptionToken = join ($self->{'resumptionseparator'}, ($set,$from,$until,$metadataPrefix,$offset+$self->{'recordlimit'}));
   }
   
   # get full count
   my $responseContainer = '';
   $sth = $self->{'dbh'}->prepare ('select count(*) '.$sqlquery);
   $sth->execute;
   if (my @row = $sth->fetchrow_array)
   {
      $responseContainer =
         "<hits ".
         "xmlns=\"$self->{'odlsearch_xmlnsprefix'}/responsecontainer\" ".
         "xsi:schemaLocation=\"$self->{'odlsearch_xmlnsprefix'}/responsecontainer ".
         "$self->{'odlsearch_xmlnsprefix'}/responsecontainer.xsd\">".
         $row[0].
         "</hits>";
   }

   ( \@allrows, $resumptionToken, $responseContainer, $metadataPrefix );
}


# forward request to ListRecords handler
sub Archive_ListIdentifiers
{
   my ($self, $set, $from, $until, $resumptionToken) = @_;

   $self->Archive_ListRecords ($set, $from, $until, '', $resumptionToken);
}


# insert a new record into the archive
sub Archive_PutRecord
{
   my ($self, $identifier, $datestamp, $metadata, $metadataPrefix,
       $metadataNamespace, $metadataSchema, $about, $status, @sets) = @_;
       
   my $sth;

   # get unique identifier if necessary
   if ($identifier eq '')
   {
      $self->{'dbh'}->do ("lock tables $self->{'table'}admin write");
      $sth = $self->{'dbh'}->prepare ("select value from $self->{'table'}admin where variable=\'counter\'");
      $sth->execute;
      if (my $rowref = $sth->fetchrow_hashref)
      {
         $identifier = "oai:$self->{'archiveId'}:$rowref->{'value'}";
      }
      $sth->finish;
      $self->{'dbh'}->do ("update $self->{'table'}admin set value=value+1 where variable=\'counter\'");
      $self->{'dbh'}->do ("unlock tables");
   }
   
   if ($status eq 'deleted')
   {
      # delete all entries
      $self->{'dbh'}->do ("update $self->{'table'} set status=\'deleted\' where identifier=\'$identifier\'");
      
      # delete all children
      $self->{'dbh'}->do ("update $self->{'table'},$self->{'table'}set set ".
                          "$self->{'table'}.status=\'deleted\' where ".
                          "$self->{'table'}set.oaiset = \'$identifier\' and".
                          "$self->{'table'}set.identifier = $self->{'table'}.identifier");
   }
   else
   {
      # search for existing entry in database
      $sth = $self->{'dbh'}->prepare ("select identifier from $self->{'table'} where identifier=\'$identifier\'");
      $sth->execute;
      my $rowref = $sth->fetchrow_hashref;
      $sth->finish;
      
      if ($rowref)
      # found the record
      {
         # update just the metadata field
         $self->{'dbh'}->do ("update $self->{'table'} set ".
                             "datestamp=\'$datestamp\', ".
                             "metadata=".$self->{'dbh'}->quote ($metadata).', '.
                             "metadataPrefix=\'$metadataPrefix\' ".
                             "where identifier=\'$identifier\'");
      }
      else
      # did not find the record
      {
         my $threadnumber = 0;
         my $level = 1;
         
         # get parent parameter
         my $parent = $sets[0];
         if ((! defined $parent) || ($parent eq ''))
         {
            $parent = 'top';
         }
         
         # find parent thread number
         $sth = $self->{'dbh'}->prepare ("select thread,level from $self->{'table'} where identifier=\'$parent\'");
         $sth->execute;
         my $rowref = $sth->fetchrow_arrayref;
         $sth->finish;
         
         if (($rowref) && ($#$rowref == 1) && ($rowref->[0] ne ''))
         # found a parent thread number
         {
            my $parentthreadnumber = $rowref->[0];
            $level = $rowref->[1] + 1;

            # find next entry in list
            $sth = $self->{'dbh'}->prepare ("select max(thread) from $self->{'table'} where thread<$parentthreadnumber");
            $sth->execute;
            my $nextthreadnumber = ($sth->fetchrow_array)[0];
            $sth->finish;
            if ($nextthreadnumber eq '')
            {
               $nextthreadnumber = 0;
            }
            
            # do a shift if necessary
            if (($nextthreadnumber + 1) >= $parentthreadnumber)
            {
               $self->{'dbh'}->do ("update $self->{'table'} set thread=thread+$self->{'threadepoch'} where thread>=$parentthreadnumber");
               $parentthreadnumber += $self->{'threadepoch'};
            }
            
            $threadnumber = int (($parentthreadnumber + $nextthreadnumber) / 2);
         }
         else
         # no parent thread number, so create new thread with large value
         {
            # search for current maximum thread number
            $sth = $self->{'dbh'}->prepare ("select max(thread) from $self->{'table'}");
            $sth->execute;
            my $maxthread = ($sth->fetchrow_array)[0];
            $sth->finish;
            
            if ($maxthread eq '')
            {
               $maxthread = 0;
            }
            $threadnumber = $maxthread + $self->{'threadepoch'};
         }

         # add new metadata record
         $self->{'dbh'}->do ("insert into $self->{'table'} set identifier=\'$identifier\', ".
                             "datestamp=\'$datestamp\', ".
                             "metadata=".$self->{'dbh'}->quote ($metadata).', '.
                             "metadataPrefix=\'$metadataPrefix\', ".
                             "level=\'$level\', ".
                             "parent=\'$parent\', ".
                             "thread=\'$threadnumber\', ".
                             "status=\'\'");

         # find sets from parent
         my @sets = ( $parent );
         $sth = $self->{'dbh'}->prepare ("select oaiset from $self->{'table'}set where identifier=\'$parent\'");
         $sth->execute;
         while (my @pset = $sth->fetchrow_array)
         {
            push (@sets, $pset[0]);
         }
         $sth->finish;

         # add sets for new entry
         foreach my $aset (@sets)
         {
            $self->{'dbh'}->do ("insert into $self->{'table'}set set identifier=\'$identifier\', oaiset=\'$aset\'");
         }
      }

      # add metadata namespace/schema
      $self->{'dbh'}->do ("delete from $self->{'table'}md where metadataPrefix=\'$metadataPrefix\' and metadataNamespace=\'$metadataNamespace\' and metadataSchema=\'$metadataSchema\'");
      $self->{'dbh'}->do ("insert into $self->{'table'}md set metadataPrefix=\'$metadataPrefix\', metadataNamespace=\'$metadataNamespace\', metadataSchema=\'$metadataSchema\'");
   }
   
   $identifier;
}


1;

